#include "pch.h"
#include "FixedBufferOscillator.h"

using namespace AudioFilterDemo;
using namespace Platform;

FixedBufferOscillator::FixedBufferOscillator(IXAudio2* pXAudio2)
{
    // Create a source voice
    WAVEFORMATEX waveFormat;
    waveFormat.wFormatTag = WAVE_FORMAT_IEEE_FLOAT;
    waveFormat.nChannels = 1;
    waveFormat.nSamplesPerSec = 48000;
    waveFormat.nAvgBytesPerSec = 48000 * 4;
    waveFormat.nBlockAlign = 4;
    waveFormat.wBitsPerSample = 32;
    waveFormat.cbSize = 0;

    HRESULT hr = pXAudio2->CreateSourceVoice(&pSourceVoice, &waveFormat, XAUDIO2_VOICE_USEFILTER, XAUDIO2_MAX_FREQ_RATIO);
    pSourceVoice->SetVolume(0);

    if (FAILED(hr))
        throw ref new COMException(hr, "CreateSourceVoice failure");

    // Submit the array
    XAUDIO2_BUFFER buffer = {0};
    buffer.AudioBytes = 4 * BUFFER_LENGTH;
    buffer.pAudioData = (byte *)waveformBuffer;
    buffer.Flags = XAUDIO2_END_OF_STREAM;
    buffer.PlayBegin = 0;
    buffer.PlayLength = BUFFER_LENGTH;
    buffer.LoopBegin = 0;
    buffer.LoopLength = BUFFER_LENGTH;
    buffer.LoopCount = XAUDIO2_LOOP_INFINITE;

    hr = pSourceVoice->SubmitSourceBuffer(&buffer);

    if (FAILED(hr))
        throw ref new COMException(hr, "SubmitSourceBuffer failure");

    // Start the voice playing
    pSourceVoice->Start();
}

void FixedBufferOscillator::SetFrequency(float freq)
{
    pSourceVoice->SetFrequencyRatio(freq / BASE_FREQ);
}

void FixedBufferOscillator::SetAmplitude(float amp)
{
    pSourceVoice->SetVolume(amp);
}

IXAudio2SourceVoice * FixedBufferOscillator::GetVoice()
{
    return pSourceVoice;
}
