﻿//
// MainPage.xaml.cpp
// Implementation of the MainPage class.
//

#include "pch.h"
#include "MainPage.xaml.h"

using namespace GraphicEqualizer;

using namespace Concurrency;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::Storage;
using namespace Windows::Storage::FileProperties;
using namespace Windows::Storage::Pickers;
using namespace Windows::Storage::Streams;
using namespace Windows::UI::Xaml;
using namespace Windows::UI::Xaml::Automation;
using namespace Windows::UI::Xaml::Controls;
using namespace Windows::UI::Xaml::Controls::Primitives;
using namespace Windows::UI::Xaml::Data;
using namespace Windows::UI::Xaml::Input;
using namespace Windows::UI::Xaml::Media;
using namespace Windows::UI::Xaml::Navigation;

MainPage::MainPage()
{
	InitializeComponent();

    // Create an IXAudio2 object
    HRESULT hresult = XAudio2Create(&pXAudio2);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "XAudio2Create failure");

    // Create a mastering voice
    hresult = pXAudio2->CreateMasteringVoice(&pMasteringVoice); 

    if (FAILED(hresult))
        throw ref new COMException(hresult, "IXAudio2->CreateMasteringVoice failure");

    // Initialize the volume
    hresult = pMasteringVoice->SetVolume(0);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "IXAudio2MasteringVoice->SetVolume failure");

    // Define constants and array for creating equalizer bands
    typedef std::tuple<float, std::wstring> freq;
    const int bandCount = 26;
    const float bandWidth = 1.0f / 3;       // in octaves

    static std::array<freq, bandCount> frequencies = 
    {
        freq(20.0f, L"20"), freq(25.0f, L"25"), freq(31.5f, L"31.5"), 
        freq(40.0f, L"40"), freq(50.0f, L"50"), freq(63.0f, L"63"),
        freq(80.0f, L"80"), freq(100.0f, L"100"), freq(125.0f, L"125"),
        freq(160.0f, L"160"), freq(200.0f, L"200"), freq(250.0f, L"250"),
        freq(320.0f, L"315"), freq(400.0f, L"400"), freq(500.0f, L"500"),
        freq(630.0f, L"630"), freq(800.0f, L"800"), freq(1000.0f, L"1K"),
        freq(1250.0f, L"1.25K"), freq(1600.0f, L"1.6K"), freq(2000.0f, L"2K"),
        freq(2500.0f, L"2.5K"), freq(3150.0f, L"3.15K"), freq(4000.0f, L"4K"),
        freq(5000.0f, L"5K"), freq(6300.0f, L"6.3K")
    };

    // Calculate Q
    float Q = float(sqrt(pow(2, bandWidth)) / (pow(2, bandWidth) - 1));

    // Create equalizer output submix voice
    IXAudio2SubmixVoice* pEqualizerOutput;
    hresult = pXAudio2->CreateSubmixVoice(&pEqualizerOutput, 2, 44100, 0, 3);

    if (FAILED(hresult))
        ref new COMException(hresult, "CreateSubmixVoice failure");

    // Set volume to compensante for filter gain
    hresult = pEqualizerOutput->SetVolume(1 / Q);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "IXAudio2SubmixVoice->SetVolume failure");

    // Create array of XAUDIO2_SEND_DESCRIPTOR structures
    XAUDIO2_SEND_DESCRIPTOR* pSendsInput = new XAUDIO2_SEND_DESCRIPTOR[bandCount];

    // Allocate memory for an array saved as a field
    pEqualizerBands = new IXAudio2SubmixVoice*[bandCount];

    // Loop through the bands
    for (int i = 0; i < bandCount; i++)
    {
        // Create a Grid column
        ColumnDefinition^ coldef = ref new ColumnDefinition();
        coldef->Width = GridLengthHelper::FromValueAndType(1, GridUnitType::Star);
        equalizerSliderGrid->ColumnDefinitions->Append(coldef);

        // Create a label
        TextBlock^ txtblk = ref new TextBlock();
        txtblk->Text = ref new String(std::get<1>(frequencies[i]).c_str());
        Grid::SetColumn(txtblk, i);
        Grid::SetRow(txtblk, 0);
        equalizerSliderGrid->Children->Append(txtblk);

        // Create a Slider
        Slider^ slider = ref new Slider();
        slider->Tag = i;
        slider->ValueChanged += ref new RangeBaseValueChangedEventHandler(this, &MainPage::OnEqualizerSliderChanged);
        Grid::SetColumn(slider, i);
        Grid::SetRow(slider, 1);
        equalizerSliderGrid->Children->Append(slider);

        // Another label
        txtblk = ref new TextBlock();
        txtblk->Text = ref new String(std::get<1>(frequencies[i]).c_str());
        Grid::SetColumn(txtblk, i);
        Grid::SetRow(txtblk, 2);
        equalizerSliderGrid->Children->Append(txtblk);

        // Define where this submix voice goes
        XAUDIO2_SEND_DESCRIPTOR sendDescriptor;
        sendDescriptor.Flags = 0;
        sendDescriptor.pOutputVoice = pEqualizerOutput;

        XAUDIO2_VOICE_SENDS sendList;
        sendList.SendCount = 1;
        sendList.pSends = &sendDescriptor;

        // Create the submix voice for that equalizer band
        hresult = pXAudio2->CreateSubmixVoice(&pEqualizerBands[i], 2, 44100, XAUDIO2_VOICE_USEFILTER, 2, &sendList);

        if (FAILED(hresult))
            throw ref new COMException(hresult, "CreateSubmixVoice failure");

        // Set the filter parameters for this band
        XAUDIO2_FILTER_PARAMETERS filterParameters;
        filterParameters.Type = XAUDIO2_FILTER_TYPE::BandPassFilter;
        float cutoffFreq = std::get<0>(frequencies[i]);
        filterParameters.Frequency = float(2 * sin(3.14159 * cutoffFreq / 44100));
        filterParameters.OneOverQ = 1 / Q;

        hresult = pEqualizerBands[i]->SetFilterParameters(&filterParameters, 0);

        if (FAILED(hresult))
            throw ref new COMException(hresult, "SetFilterParameters failure");

        // Define an XAUDIO2_SEND_DESCRIPTOR for the graphic equalizer input
        sendDescriptor.Flags = 0;
        sendDescriptor.pOutputVoice = pEqualizerBands[i];
        pSendsInput[i] = sendDescriptor;
    }

    // Create another submix voice for the equalizer output
    XAUDIO2_VOICE_SENDS sendListInput;
    sendListInput.SendCount = bandCount;
    sendListInput.pSends = pSendsInput;
    hresult = pXAudio2->CreateSubmixVoice(&pEqualizerInput, 2, 44100, 0, 1, &sendListInput);

    if (FAILED(hresult))
        throw ref new COMException(hresult, "CreateSubmixVoice failure");

    // No longer need this allocated array
    delete[] pSendsInput;

    // Create submix voice to connect source voice with equalizer
    hresult = pXAudio2->CreateSubmixVoice(&pSourceVoiceOutput, 2, 44100, 0, 0);

    if (FAILED(hresult))
        ref new COMException(hresult, "CreateSubmixVoice failure");

    // And attach it
    AttachSourceVoiceToEqualizer();
}

void MainPage::OnOpenButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    FileOpenPicker^ fileOpenPicker = ref new FileOpenPicker();
    fileOpenPicker->SuggestedStartLocation = PickerLocationId::MusicLibrary;
    fileOpenPicker->FileTypeFilter->Append(".wma");
    fileOpenPicker->FileTypeFilter->Append(".mp3");
    fileOpenPicker->FileTypeFilter->Append(".wav");

    // Get a music file from the picker
    task<StorageFile^> pickSingleFileTask = create_task(fileOpenPicker->PickSingleFileAsync());
    pickSingleFileTask.then([this](StorageFile^ storageFile)
    {
        if (storageFile == nullptr)
            return;

        // Get the thumbnail for the display 
        task<StorageItemThumbnail^> getThumbnailTask = create_task(storageFile->GetThumbnailAsync(ThumbnailMode::SingleItem));
        getThumbnailTask.then([this](StorageItemThumbnail^ storageItemThumbnail)
        {
            task<void> setSourceTask(bitmapImage->SetSourceAsync(storageItemThumbnail));
            setSourceTask;
        });

        // Get information about the track
        StorageItemContentProperties^ contentProperties = storageFile->Properties;
        task<MusicProperties^> getMusicPropertiesTask = create_task(contentProperties->GetMusicPropertiesAsync());
        getMusicPropertiesTask.then([this](MusicProperties^ musicProperties)
        {
            artistText->Text = musicProperties->Artist;
            albumText->Text = musicProperties->Album;
            trackText->Text = musicProperties->Title;

            int durationInSeconds = int(musicProperties->Duration.Duration / 10000000);
            wchar_t str[16];
            swprintf(str, 16, L"%i:%02i", durationInSeconds / 60, 
                                          durationInSeconds % 60);
            durationText->Text = ref new String(str);
        });

        // Open the file for reading
        task<IRandomAccessStream^> openReadTask = create_task(storageFile->OpenAsync(FileAccessMode::Read));
        openReadTask.then([this](IRandomAccessStream^ randomAccessStream)
        {
            // Create an IMFSourceReader and an IXAudio2SourceVoice 
            if (pAudioFilePlayer != nullptr)
                delete pAudioFilePlayer;

            pAudioFilePlayer = new AudioFilePlayer(pXAudio2, pSourceVoiceOutput, randomAccessStream);

            // Quite possibly start it playing
            if (pauseButton->Visibility == Windows::UI::Xaml::Visibility::Visible)
                pAudioFilePlayer->Start();
        });
    });
}

void GraphicEqualizer::MainPage::OnPlayButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    if (pAudioFilePlayer != nullptr)
        pAudioFilePlayer->Start();

    playButton->Visibility = Windows::UI::Xaml::Visibility::Collapsed;
    pauseButton->Visibility= Windows::UI::Xaml::Visibility::Visible;
}

void GraphicEqualizer::MainPage::OnPauseButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    if (pAudioFilePlayer != nullptr)
        pAudioFilePlayer->Stop();

    playButton->Visibility = Windows::UI::Xaml::Visibility::Visible;
    pauseButton->Visibility= Windows::UI::Xaml::Visibility::Collapsed;
}

void MainPage::OnVolumeSliderValueChanged(Object^ sender, RangeBaseValueChangedEventArgs^ args)
{
    HRESULT hresult = pMasteringVoice->SetVolume(float(args->NewValue / 100));

    if (FAILED(hresult))
        ref new COMException(hresult, "IXAudio2MasteringVoice->SetVolume failure");
}

void MainPage::OnEqualizerInOutButtonClick(Object^ sender, RoutedEventArgs^ args)
{
    CheckBox^ checkBox = dynamic_cast<CheckBox^>(sender);

    if (checkBox->IsChecked->Value)
    {
        AttachSourceVoiceToEqualizer();
        AutomationProperties::SetName(checkBox, "Equalizer In");
    }
    else
    {
        pSourceVoiceOutput->SetOutputVoices(nullptr);
        AutomationProperties::SetName(checkBox, "Bypass Equalizer");
    }
}

void MainPage::AttachSourceVoiceToEqualizer()
{
    XAUDIO2_SEND_DESCRIPTOR sendDescriptor;
    sendDescriptor.Flags = 0;
    sendDescriptor.pOutputVoice = pEqualizerInput;

    XAUDIO2_VOICE_SENDS sendList;
    sendList.SendCount = 1;
    sendList.pSends = &sendDescriptor;

    pSourceVoiceOutput->SetOutputVoices(&sendList);
}

void MainPage::OnEqualizerSliderChanged(Platform::Object^ sender, RangeBaseValueChangedEventArgs^ args)
{
    Slider^ slider = dynamic_cast<Slider^>(sender);
    int bandIndex = (int)slider->Tag;
    float amplitude = float(pow(10, (args->NewValue / 20)));
    pEqualizerBands[bandIndex]->SetVolume(amplitude, 0);
}

